#!/usr/bin/env lua

-- Lua script to convert output of Radford M. Neil's "Software for Low Density
-- Parity Check (LDPC) codes", release of 2012-02-11, to "check.c" and
-- "generator.c" code files.
--
-- Radford Neil's software for LDPC codes is not part of this software package
-- and neither a compile nor a run-time dependency if you use the included
-- "check.c" and "generator.c" files.
--
-- As of 2019-02-12, Radford Neil's software is available at:
-- https://www.cs.utoronto.ca/~radford/ldpc.software.html

ldpc_prog_path, c, n, options = ...
c = tonumber(c)
n = tonumber(n)
os.execute("rm -f ldpc.pchk ldpc.gen ldpc.pchk.txt ldpc.gen.txt")
os.execute(ldpc_prog_path .. "/make-ldpc ldpc.pchk " .. c .. " " .. n .. " " .. options)
os.execute(ldpc_prog_path .. "/make-gen ldpc.pchk ldpc.gen dense")
os.execute(ldpc_prog_path .. "/print-pchk ldpc.pchk > ldpc.pchk.txt")
os.execute(ldpc_prog_path .. "/print-gen ldpc.gen > ldpc.gen.txt")

pchk = {}
gen = {}
gen_row_cnt = 0
order = {}
order_cnt = 0
mode = nil
for line in io.lines("ldpc.gen.txt") do
  if line == "Column order (message bits at end):" then
    mode = "order"
  elseif line == "Inv(A) X B:" then
    mode = "gen"
  elseif line:find("^ ") then
    if mode == "order" then
      for col in line:gmatch("[0-9]+") do
        order[tonumber(col)] = order_cnt
        order_cnt = order_cnt + 1
      end
    elseif mode == "gen" then
      local row, col_cnt = {}, 0
      for entry in line:gmatch("[01]") do
        row[col_cnt] = tonumber(entry)
        col_cnt = col_cnt + 1
      end
      gen[gen_row_cnt] = row
      gen_row_cnt = gen_row_cnt + 1
    end
  end
end
for line in io.lines("ldpc.pchk.txt") do
  local row_idx, rest = line:match("^ *([0-9]+):(.*)")
  if row_idx then
    local row = {}
    for col in rest:gmatch("[0-9]+") do
      row[#row+1] = (order[tonumber(col)] + (n-c)) % n
    end
    table.sort(row)
    pchk[tonumber(row_idx)] = row
  end
end

bits = {}
for bit = 0, n-1 do
  local row = {}
  for row_idx = 0, c-1 do
    for entry_idx, entry in ipairs(pchk[row_idx]) do
      if entry == bit then
        row[#row+1] = row_idx
        break
      end
    end
  end
  bits[bit] = row
end

file_gen = assert(io.open("generator.c", "w"))
file_gen:write('const uint8_t pktfec_generator[] = "')
local mask = 0x80
local val = 0
for col = 0, (n-c)-1 do
  for row = 0, c-1 do
    if gen[row][col] == 1 then val = val | mask end
    mask = mask >> 1
    if mask == 0 then
      file_gen:write(string.format("\\x%02x", val))
      mask = 0x80
      val = 0
    end
  end
end
file_gen:write('";\n')
file_gen:close()

file_pchk = assert(io.open("check.c", "w"))

function write_pchk(name, pchk, rowcnt)
  local maxcnt = 0
  for row_idx = 0, rowcnt-1 do
    maxcnt = math.max(maxcnt, #pchk[row_idx])
  end
  file_pchk:write("const int16_t " .. name .. "[" .. rowcnt .. "][" .. (maxcnt+1) .. "] = {\n")
  for row_idx = 0, rowcnt-1 do
    if row_idx > 0 then
      file_pchk:write(",\n")
    end
    file_pchk:write("  {")
    for entry_idx, entry in ipairs(pchk[row_idx]) do
      file_pchk:write(entry .. ", ")
    end
    file_pchk:write("-1}")
  end
  file_pchk:write("\n};\n");
end

write_pchk("pktfec_checks", pchk, c)
write_pchk("pktfec_bitchecks", bits, n)


file_pchk:close();

