#include "pktfec.h"
#include "check.c"

#include <math.h>

#include <stdio.h>
static bool pktfec_decode_block(uint8_t *output, float *input) {
  float probs[8*PKTFEC_BLOCKLEN];
  int round, i, j, k;
  // iterative solver
  for (round=0; round<PKTFEC_ITERATIONS; round++) {
    // check if all parity checks succeed
    for (i=0; i<8*PKTFEC_BLOCKFECLEN; i++) {
      const int16_t *check = pktfec_checks[i];
      float val = 1.0;
      for (j=0; check[j]>=0; j++) val *= input[check[j]];
      if (!(val > 0.0)) goto pktfec_decode_block_continue;
    }
    // all parity checks succeeded
    for (i=0; i<PKTFEC_BLOCKLEN; i++) {
      uint8_t byte = 0, mask;
      for (mask=1<<7; mask; mask>>=1) if (*input++ < 0) byte |= mask;
      *output++ = byte;
    }
    return true;
    // not all parity checks succeeded
    pktfec_decode_block_continue:
    // convert floats from -inf to inf into probability values where +1.0 is a
    // guaranteed 0 bit, -1.0 is a guaranteed 1 bit, and 0.0 means unknown
    for (i=0; i<8*PKTFEC_BLOCKLEN; i++) probs[i] = tanhf(input[i]);
    // correct each bit
    for (i=0; i<8*PKTFEC_BLOCKLEN; i++) {
      float corr = 0.0;
      // use all parity equations containing bit to be corrected
      for (j=0; pktfec_bitchecks[i][j] >= 0; j++) {
        const int16_t *check = pktfec_checks[pktfec_bitchecks[i][j]];
        // calculate bit probability based on other bits in equation
        float est = 1.0;
        for (k=0; check[k]>=0; k++) if (i!=check[k]) est *= probs[check[k]];
        // add to float value atanh(est) == (ln(P0)-ln(P1))/2 where P0 is
        // probability for bit=0 and P1 is proability for bit=1
        corr += atanhf(est);
      }
      input[i] += (float)PKTFEC_SPEED * corr;
    }
  }
  return false;
}

bool pktfec_decode_start(
  pktfec_decoder_t *decoder, const float *input, size_t length
) {
  int i;
  float inbuf[8*PKTFEC_BLOCKLEN];
  size_t datalen;
  // read and decode first block
  for (i=0; i<8*PKTFEC_BLOCKLEN; i++) inbuf[i] = (i<length) ? input[i] : 0.0;
  if (!pktfec_decode_block(decoder->firstdata, inbuf)) return false;
  // determine payload length and length of length information
  if (decoder->firstdata[PKTFEC_BLOCKDATALEN-1] < 0x80) {
    decoder->outputlen = decoder->firstdata[PKTFEC_BLOCKDATALEN-1];
    decoder->lenlen = 1;
  } else if (decoder->firstdata[PKTFEC_BLOCKDATALEN-1] < 0xc0) {
    decoder->outputlen = 0x80 + (
      ((decoder->firstdata[PKTFEC_BLOCKDATALEN-1] & 0x3f) << 8) |
      decoder->firstdata[PKTFEC_BLOCKDATALEN-2]
    );
    decoder->lenlen = 2;
  } else if (decoder->firstdata[PKTFEC_BLOCKDATALEN-1] < 0xe0) {
    decoder->outputlen = 0x4080 + (
      ((decoder->firstdata[PKTFEC_BLOCKDATALEN-1] & 0x1f) << 16) |
      (decoder->firstdata[PKTFEC_BLOCKDATALEN-2] << 8) |
      decoder->firstdata[PKTFEC_BLOCKDATALEN-3]
    );
    decoder->lenlen = 3;
  } else {
    // unsupported length
    return false;
  }
  // calculate total length of payload for LDPC FEC algorithm
  datalen = decoder->lenlen + decoder->outputlen + PKTFEC_CRCLEN;
  // calculate total number of blocks needed
  decoder->blockcount = (datalen-1) / PKTFEC_BLOCKDATALEN + 1;
  // if further input data is needed
  if (decoder->blockcount > 1) {
    // calculate payload and parity size of last block
    decoder->lastblockdatalen = (datalen-1) % PKTFEC_BLOCKDATALEN + 1;
    decoder->lastblockfeclen = PKTFEC_SHORTBLOCKFECLEN(
      decoder->lastblockdatalen
    );
    // calculate theoretic/full input length that goes beyond one block
    decoder->extralen = (
      (decoder->blockcount-2) * PKTFEC_BLOCKLEN +
      decoder->lastblockdatalen + decoder->lastblockfeclen
    );
  } else {
    // set extralen to zero to indicate no further input data is needed
    decoder->extralen = 0;
  }
  // report success
  return true;
}

bool pktfec_decode_finish(
  pktfec_decoder_t *decoder, uint8_t *output, const float *input, size_t length
) {
  uint8_t outbuf[PKTFEC_BLOCKLEN];
  float inbuf[8*PKTFEC_BLOCKLEN];
  size_t inputidx = 0, outputidx = 0;
  int blockidx=0, byteidx, bitidx;
  int blockdatalen, blockfeclen;
  uint32_t crc = 0xffffffff;
  int shuffleA, shuffleB, shuffleC;
  // restore first block
  for (byteidx=0; byteidx<PKTFEC_BLOCKDATALEN; byteidx++) {
    outbuf[byteidx] = decoder->firstdata[byteidx];
  }
  // set payload and parity size for first block
  blockdatalen = PKTFEC_BLOCKDATALEN - decoder->lenlen;
  blockfeclen = PKTFEC_BLOCKFECLEN;
  // if payload including CRC is smaller than payload area of first block,
  // then set payload length for FEC to net payload length plus CRC size
  if (blockdatalen > decoder->outputlen + PKTFEC_CRCLEN) {
    blockdatalen = decoder->outputlen + PKTFEC_CRCLEN;
  }
  while (1) {
    // process blockdatalen bytes
    for (byteidx=0; byteidx<blockdatalen; byteidx++) {
      if (outputidx < decoder->outputlen) {
        // process payload
        if (output) output[outputidx++] = outbuf[byteidx];
        // calculate CRC with most significant bit first
        crc ^= (uint32_t)outbuf[byteidx] << 24;
        for (bitidx=0; bitidx<8; bitidx++) {
          crc = (crc<<1) ^ ((-((crc&(1<<31))>>31)) & PKTFEC_CRC32POLY);
        }
      } else {
        // payload done, check CRC
        if (outbuf[byteidx] != (uint8_t)~(crc >> 24)) return false;
        crc <<= 8;
      }
    }
    // switch to next block if existent
    if (++blockidx==decoder->blockcount) break;
    // set blockdatalen and blockfeclen
    if (blockidx == decoder->blockcount-1) {
      blockdatalen = decoder->lastblockdatalen;
      blockfeclen = decoder->lastblockfeclen;
    } else {
      blockdatalen = PKTFEC_BLOCKDATALEN;
      blockfeclen = PKTFEC_BLOCKFECLEN;
    }
    // fill implicit data area of last block if applicable
    for (byteidx=blockdatalen; byteidx<PKTFEC_BLOCKDATALEN; byteidx++) {
      for (bitidx=0; bitidx<8; bitidx++) {
        inbuf[8*byteidx+bitidx] = 1.0/0.0;
      }
    }
    // clear implicit FEC area of last block if applicable
    for (byteidx=blockfeclen; byteidx<PKTFEC_BLOCKFECLEN; byteidx++) {
      for (bitidx=0; bitidx<8; bitidx++) {
        inbuf[8*PKTFEC_BLOCKDATALEN+8*byteidx+bitidx] = 0.0;
      }
    }
    // read block while deshuffling
    inputidx = blockidx - 1;
    shuffleA = 0; shuffleB = -1;
    for (byteidx=0; byteidx<blockdatalen+blockfeclen; byteidx++) {
      if (blockidx == decoder->blockcount-1) {
        while (1) {
          shuffleC = (
            shuffleA++ *
            (decoder->lastblockdatalen+decoder->lastblockfeclen) /
            PKTFEC_BLOCKLEN
          );
          if (shuffleB != shuffleC) {
            shuffleB = shuffleC;
            break;
          }
          inputidx += decoder->blockcount-2;
        }
      }
      for (bitidx=0; bitidx<8; bitidx++) {
        inbuf[
          8*(
            byteidx < blockdatalen ?
            byteidx :
            PKTFEC_BLOCKDATALEN + byteidx - blockdatalen
          ) + bitidx
        ] = (
          8*inputidx+bitidx < length ?
          input[8*inputidx+bitidx] :
          0.0f
        );
      }
      if (blockidx == decoder->blockcount-1) {
        inputidx += decoder->blockcount-1;
      } else {
        inputidx += decoder->blockcount-2;
        shuffleC = (
          shuffleA++ *
          (decoder->lastblockdatalen+decoder->lastblockfeclen) /
          PKTFEC_BLOCKLEN
        );
        if (shuffleB != shuffleC) {
          shuffleB = shuffleC;
          inputidx++;
        }
      }
    }
    // decode block
    if (!pktfec_decode_block(outbuf, inbuf)) return false;
  }
  // report success
  return true;
}

