#include "qam.h"

#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>

static void qam_switch(qam_modulator_t *state, int I, int Q) {
  int sample;
  float phase, progress;
  float *const fout = (float *)(state->output + state->output_wpos);
  for (sample=0; sample<state->config->switch_samples; sample++) {
    phase = state->config->carrier_phase_step * sample;
    progress = (
      0.5f * (1.0f-cosf((sample+0.5f) * state->config->switch_phase_step))
    );
    fout[sample] = state->config->tx_amplitude * (
      (state->I*(1.0f-progress)+I*progress) * sinf(phase) * state->phase_odd +
      (state->Q*(1.0f-progress)+Q*progress) * cosf(phase) * state->phase_odd
    );
  }
  state->phase_odd *= state->config->switch_odd;
  state->I = I;
  state->Q = Q;
  state->output_wpos += state->config->switch_samples * sizeof(float);
}

static void qam_hold(qam_modulator_t *state) {
  int sample;
  float phase;
  float *const fout = (float *)(state->output + state->output_wpos);
  for (sample=0; sample<state->config->hold_samples; sample++) {
    phase = state->config->carrier_phase_step * sample;
    fout[sample] = state->config->tx_amplitude * (
      state->I * sinf(phase) * state->phase_odd +
      state->Q * cosf(phase) * state->phase_odd
    );
  }
  state->phase_odd *= state->config->hold_odd;
  state->output_wpos += state->config->hold_samples * sizeof(float);
}

static unsigned int qam_grayencode(unsigned int n) {
  unsigned int mask = n;
  while (mask >>= 1) n ^= mask;
  return n;
}

static void qam_writesymbol_nosync(qam_modulator_t *state, unsigned int symbol) {
  qam_switch(
    state,
    state->config->max_amplitude_factor - 2 * qam_grayencode(
      symbol >> (state->config->bits_per_symbol/2)
    ),
    state->config->max_amplitude_factor - 2 * qam_grayencode(
      symbol & ((1<<(state->config->bits_per_symbol/2))-1)
    )
  );
  qam_hold(state);
}

static void qam_writesymbol_withsync(qam_modulator_t *state) {
  state->sync = (state->sync + 1) % state->config->qamsync_resync_interval;
  if (!state->sync) {
    qam_writesymbol_nosync(state, 0);
    state->sync = 1;
  }
  qam_writesymbol_nosync(state, state->symbol);
}

void qam_modulator_symsync(qam_modulator_t *state) {
  qam_switch(
    state,
    -state->config->max_amplitude_factor,
    -state->config->max_amplitude_factor
  );
  qam_hold(state);
  qam_switch(
    state,
    state->config->max_amplitude_factor,
    state->config->max_amplitude_factor
  );
  qam_hold(state);
}

static void qam_modulator_init(
  qam_modulator_t *state,
  qam_config_t *config,
  qam_writer_fptr writer,
  void *writer_ctx
) {
  void *tmp = state->output;
  memset(state, 0, sizeof(qam_modulator_t));
  state->output = tmp;
  state->config = config;
  state->writer = writer;
  state->writer_ctx = writer_ctx;
  state->phase_odd = 1;
  qam_modulator_symsync(state);
}

qam_modulator_t *qam_modulator_create(
  qam_config_t *config,
  qam_writer_fptr writer,
  void *writer_ctx
) {
  qam_modulator_t *state;
  if (!config->valid) return NULL;
  state = malloc(sizeof(qam_modulator_t));
  if (!state) return NULL;
  state->output = malloc(2 * config->symbol_samples * sizeof(float));
  if (!state->output) {
    free(state);
    return NULL;
  }
  qam_modulator_init(state, config, writer, writer_ctx);
  return state;
}

void qam_modulator_reset(qam_modulator_t *state) {
  qam_modulator_init(state, state->config, state->writer, state->writer_ctx);
}

void qam_modulator_destroy(qam_modulator_t *state) {
  free(state->output);
  free(state);
}

void qam_modulator_send(qam_modulator_t *state, uint8_t *data, size_t length) {
  if (state->input_wpos || state->input_finish) abort();
  state->input = data;
  state->input_wpos = length;
}

void qam_modulator_finish(qam_modulator_t *state) {
  if (state->input_finish > 1) abort();
  state->input_finish = 1;
}

bool qam_modulator_run(qam_modulator_t *state) {
  while (1) {
    if (state->output_rpos < state->output_wpos) {
      state->output_rpos += state->writer(
        state->writer_ctx,
        state->output + state->output_rpos,
        state->output_wpos - state->output_rpos
      );
    }
    if (state->output_rpos < state->output_wpos) return false;
    state->output_rpos = 0;
    if (state->stage == QAM_STAGE_SYMSYNC) {
      if (!state->sync) {
        state->output_wpos = 0;
        qam_modulator_symsync(state);
        state->sync = state->config->symsync_tx_half_count-1;
      } else if (!--state->sync) {
        state->output_wpos = 0;
        qam_hold(state);
        qam_hold(state);
        state->stage = QAM_STAGE_QAMSYNC;
        state->sync = state->config->qamsync_symbol_count/2;
      }
    } else if (state->stage == QAM_STAGE_QAMSYNC) {
      if (!--state->sync) {
        state->output_wpos = 0;
        state->stage = QAM_STAGE_SYNCED;
      }
    } else {
      state->output_wpos = 0;
      if (
        state->bitcount >= state->config->bits_per_symbol || (
          state->bitcount > 0 &&
          state->input_rpos == state->input_wpos &&
          state->input_finish == 1
        )
      ) {
        if (state->bitcount >= state->config->bits_per_symbol) {
          state->symbol = state->bitbuffer >> (
            state->bitcount - state->config->bits_per_symbol
          );
        } else {
          state->symbol = state->bitbuffer << (
            state->config->bits_per_symbol - state->bitcount
          );
        }
        qam_writesymbol_withsync(state);
        state->bitcount -= state->config->bits_per_symbol;
        state->bitbuffer &= (1<<state->bitcount) - 1;
      } else if (state->input_rpos < state->input_wpos) {
        state->bitbuffer <<= 8;
        state->bitbuffer |= state->input[state->input_rpos++];
        state->bitcount += 8;
      } else if (state->input_finish == 1) {
        qam_switch(state, 0, 0);
        state->input_finish = 2;
      } else {
        state->input_rpos = 0;
        state->input_wpos = 0;
        return true;
      }
    }
  }
}

typedef struct {
  size_t length;
  size_t bufsize;
  uint8_t *data;
  int memerror;
} qam_modulate_outbuf_t;

size_t qam_modulate_writer(
  void *ctx, uint8_t *output, size_t length
) {
  qam_modulate_outbuf_t *const outbuf = ctx;
  if (outbuf->memerror) return 0;
  if (length > outbuf->bufsize - outbuf->length) {
    while (length > outbuf->bufsize - outbuf->length) {
      if (outbuf->bufsize > (SIZE_MAX >> 1)) {
        free(outbuf->data);
        outbuf->memerror = 1;
        return 0;
      }
      outbuf->bufsize <<= 1;
    }
    {
      uint8_t *newptr = realloc(outbuf->data, outbuf->bufsize);
      if (!newptr) {
        free(outbuf->data);
        outbuf->memerror = 1;
        return 0;
      }
      outbuf->data = newptr;
    }
  }
  {
    size_t i = length;
    while (i--) outbuf->data[outbuf->length++] = *output++;
  }
  return length;
}

size_t qam_modulate(
  qam_config_t *config,
  float **output, uint8_t *data, size_t length
) {
  qam_modulator_t *state;
  qam_modulate_outbuf_t outbuf;
  outbuf.length = 0;
  outbuf.bufsize = 1024;
  outbuf.data = malloc(outbuf.bufsize);
  if (!outbuf.data) { *output = NULL; return 0; }
  outbuf.memerror = 0;
  state = qam_modulator_create(config, qam_modulate_writer, &outbuf);
  if (!state) { free(outbuf.data); *output = NULL; return 0; }
  qam_modulator_send(state, data, length);
  qam_modulator_finish(state);
  while (!qam_modulator_run(state) && !outbuf.memerror) {}
  qam_modulator_destroy(state);
  if (outbuf.memerror) { *output = NULL; return 0; }
  *output = (float *)outbuf.data;
  return outbuf.length / sizeof(float);
}

